/* Instruction.java -	(c) Ilpo Jrvinen
 *
 * TiTo K2000 Harj.
 */


import	java.util.*;

public class Instruction {

	public static final int UNKNOWN_OPCODE = -1;
	
	private int	i;

	public Instruction( int i ) {
		this.i = i;
	}
    
	public int getValue()		{ return i; }
	public int getRj()			{ return (i >> 21) & 0x07; }
	public int getM()			{ return (i >> 19) & 0x03; }
	public int getRi()			{ return (i >> 16) & 0x07; }
	public int getAddr()		{ return (i & 0x7fff) * getSign(); }		// handle < 0 too
	public int getSign()		{ return ((i & 0x8000) != 0) ? -1 : 1; }
	public int getAddrField()	{ return (i >> 00) & 0xffff; }
            
	public int getOpField() { return (i >> 24) & 0xff; }
	public int getOp() 
	{
		int op = (i >> 24) & 0xff;
		if ((op > 0x04 && op < 0x11) || (op > 0x1a && op < 0x1f) ||
			(op > 0x2c && op < 0x31) || (op > 0x34 && op < 0x70) || (op > 0x70)) {   
			return UNKNOWN_OPCODE;
		}
		return op;
	}
	
	public int getEA( Cpu cpu )
	{
		return getAddr() + (getRi() != 0 ? cpu.from( getRi() ) : 0 );
	}
	
	public int getData( Cpu cpu )
	{
		int addr = getEA( cpu ), count = getM();
		if (count > 2)
			cpu.halt( "Invalid addressing mode" );
		for (int i = 0; i < count; i++)
			addr = cpu.getMem( addr );		// all addressing modes
		return addr;
	}
	
//	public int getAddr( Cpu cpu )
//	{
//		return getData( cpu );
//	}
	
	public boolean isStore() 
	{
		int op = getOp();
		return op == STORE || op == JUMP || op == JZER;		// Is this correct?
	}
	
	public String getSymbolic() 
	{
		if (!ops.containsKey(new Integer(getOp()))) {
			return "NOT IMPLEMENTED";
		}
		return (String)ops.get(new Integer(getOp())) + " R" + getRj() + ", " +
				getAddrSymbolic();
	}
	
	public String getAddrSymbolic()
	{
		int count = getM();
		if (isStore())
			count++;
		switch( count ) {
			case 0:
				if (getAddr() == 0 && getRi() != 0) {
					return "R" + getRi();
				}
				return "=" + getAddr() + (getRi() != 0 ? "(R" + getRi() + ")" : "");
			case 1:
				return getAddr() + (getRi() != 0 ? "(R" + getRi() + ")" : "");
			case 2:
				return "@" + getAddr() + (getRi() != 0 ? "(R" + getRi() + ")" : "");
			case 3: throw new RuntimeException( "Invalid addressing mode" );
			default: throw new RuntimeException( "BUG? m > 3?" );
		}
	}
	
	
	public String toString() 
	{
		return i + "\t" + (i < 10000000 ? "\t" : "" ) +
					(getOpField() < 100 ? " " : "" ) +
					(getOpField() < 10 ? " " : "" ) +
					getOpField() + ":" + getRj() + ":" +
					getM() + ":" + getRi() + ":" + getAddrField() + "\t" +
					getSymbolic();
	}



	/*
	 * Opcode initalization...
	 */
	public static final int NOP		= 0x00;
	public static final int STORE	= 0x01;
	public static final int LOAD	= 0x02;
	public static final int ADD		= 0x11;
	public static final int SUB		= 0x12;
	public static final int MUL		= 0x13;
	public static final int DIV		= 0x14;
	public static final int MOD		= 0x15;
	public static final int COMP	= 0x1f;
	public static final int JUMP	= 0x20;	
	public static final int JZER	= 0x22;
	public static final int JLES	= 0x27;
	public static final int JEQU	= 0x28;
	public static final int JGRE	= 0x29;
	public static final int CALL	= 0x31;
	public static final int EXIT	= 0x32;
	public static final int PUSH	= 0x33;
	public static final int POP		= 0x34;
	public static final int SVC		= 0x70;
	public static final int UNKNOWN = -1;
	
	public static final int HALT	= 11;
	
	private static Hashtable ops = new Hashtable();
	
	static {
		ops.put(new Integer( NOP ), new String( "NOP" ) );
		ops.put(new Integer( STORE ), new String( "STORE" ) );
		ops.put(new Integer( LOAD ), new String( "LOAD" ) );
		ops.put(new Integer( ADD ), new String( "ADD" ) );
		ops.put(new Integer( SUB ), new String( "SUB" ) );
		ops.put(new Integer( MUL ), new String( "MUL" ) );
		ops.put(new Integer( DIV ), new String( "DIV" ) );
		ops.put(new Integer( MOD ), new String( "MOD" ) );
		ops.put(new Integer( COMP ), new String( "COMP" ) );
		ops.put(new Integer( JUMP ), new String( "JUMP" ) );
		ops.put(new Integer( JZER ), new String( "JZER" ) );
		ops.put(new Integer( JLES ), new String( "JLES" ) );
		ops.put(new Integer( JEQU ), new String( "JEQU" ) );
		ops.put(new Integer( JGRE ), new String( "JGRE" ) );
		ops.put(new Integer( CALL ), new String( "CALL" ) );
		ops.put(new Integer( EXIT ), new String( "EXIT" ) );
		ops.put(new Integer( PUSH ), new String( "PUSH" ) );
		ops.put(new Integer( POP ), new String( "POP" ) );
		ops.put(new Integer( SVC ), new String( "SVC" ) );
		ops.put(new Integer( UNKNOWN ), new String( "---" ) );
	
	}
	
}